import enum
import uuid
import secrets

from sqlalchemy import Column, String, Integer, ForeignKey, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.sql.functions import now
from sqlalchemy_utils.types.choice import ChoiceType
from apistar_sqlalchemy import database

from .guid import GUID
from .hasher import hasher


class UserRole(enum.Enum):
    admin = 1
    user = 2


def can_user_create_user(user, new_user):
    if not user:
        # Unauthenticated users can only create a new user account.
        return new_user.role == UserRole.user

    # Admins can create any account.
    if user.role == UserRole.admin:
        return True

    # Else, users can only create a user account.
    assert user.role == UserRole.user
    return user.role == new_user.role


class User(database.Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    username = Column(String, nullable=False, unique=True)
    password = Column(String, nullable=False)
    role = Column(ChoiceType(UserRole, impl=Integer()), nullable=False)
    fullname = Column(String)

    created = Column(DateTime(timezone=True), server_default=now())
    updated = Column(DateTime(timezone=True), server_default=now(),
                     onupdate=now())

    sessions = relationship('UserSession',  # order_by='user_sessions.created',
                            back_populates='user')

    def __init__(self, *args, **kwargs):
        # Convert role name to enum value.
        kwargs['role'] = UserRole[kwargs['role']]
        super().__init__(*args, **kwargs)
        self.password = hasher().encrypt(self.password)

    def verify_password(self, password):
        return hasher().verify(password, self.password)

    def __repr__(self):
        msg = '<User(username=%r, role=%s, fullname=%r)>'
        return msg % (self.username, self.role.name, self.fullname)


class UserSession(database.Base):
    __tablename__ = 'user_sessions'

    id = Column(GUID, primary_key=True)
    user_id = Column(Integer, ForeignKey('users.id'))

    created = Column(DateTime(timezone=True), server_default=now())
    updated = Column(DateTime(timezone=True), server_default=now(),
                     onupdate=now(), index=True)

    user = relationship('User', back_populates='sessions')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = self.generate_session_id()

    def __repr__(self):
        msg = '<UserSession(id=%r, user_id=%r, created=%s, updated=%s)>'
        return msg % (self.id, self.user_id, self.created, self.updated)

    def generate_session_id(self):
        return uuid.UUID(bytes=secrets.token_bytes(16))


class Token(database.Base):
    __tablename__ = 'tokens'

    id = Column(GUID, primary_key=True)
    user_id = Column(Integer, ForeignKey('users.id'))

    created = Column(DateTime(timezone=True), server_default=now())
    updated = Column(DateTime(timezone=True), server_default=now(),
                     onupdate=now(), index=True)

    user = relationship('User')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = self.generate_session_id()

    def __repr__(self):
        msg = '<Token(id=%r, user_id=%r, created=%s, updated=%s)>'
        return msg % (self.id, self.user_id, self.created, self.updated)

    def generate_session_id(self):
        return uuid.UUID(bytes=secrets.token_bytes(16))



import uuid
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID


class GUID(TypeDecorator):
    '''Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    '''
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if dialect.name == 'postgresql':
            return str(value)

        if not isinstance(value, uuid.UUID):
            return "%.32x" % uuid.UUID(value).int
        # hexstring
        return "%.32x" % value.int

    def process_literal_param(self, value, dialect):
        raise NotImplementedError()

    def process_result_value(self, value, dialect):
        if value is None:
            return value

        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        return value

    def python_type(self):
        raise NotImplementedError()



from typing import List

from apistar import Route, validators, types, http
from sqlalchemy.orm import Session

from .auth import authorized
from .models import Token, User
from .validators import UUID


class TokenType(types.Type):
    id = UUID()
    user_id = validators.Integer()

    created = validators.DateTime()
    updated = validators.DateTime()


@authorized
def list_tokens(session: Session, user: User) -> List[TokenType]:
    tokens = session.query(Token).filter(Token.user_id == user.id).all()
    return list(map(TokenType, tokens))


@authorized
def create_token(session: Session, user: User) -> http.JSONResponse:
    token = Token(user=user)
    session.add(token)
    session.commit()
    return http.JSONResponse(TokenType(token), status_code=201)


routes = [
    Route('/', 'GET', list_tokens),
    Route('/', 'POST', create_token),
]



# -*- coding: utf-8 -*-

from apistar import Include

from .auth import (
    AuthorizationHook,
    authorized,
    Unauthorized,
)

from .cookies import SESSION_COOKIE_NAME

from .models import (
    Token,
    User,
    UserRole,
    UserSession,
)

from .login import (
    routes as login_routes
)

from .tokens import (
    routes as tokens_routes,
)

from .users import (
    routes as users_routes,
    UserInputType,
    UserType,
    UserComponent,
)

from .hasher import (
    disable_bcrypt_hasher,
    enable_bcrypt_hasher,
)

__version__ = '0.3.3'
__license__ = 'MIT'

__author__ = 'Sander Mathijs van Veen'
__email__ = 'sandervv+pypi@gmail.com'

__url__ = 'https://github.com/smvv/apistar-auth'
__description__ = \
    'Authentication integration based on SQLAlchemy for API Star.'

__all__ = [
    'AuthorizationHook', 'authorized', 'Unauthorized',
    'SESSION_COOKIE_NAME',
    'disable_bcrypt_hasher', 'enable_bcrypt_hasher',
    'Token',
    'User', 'UserInputType', 'UserType', 'UserRole', 'UserSession',
    'UserComponent',
    'routes',
]

routes = [
    Include('/login', name='login', routes=login_routes),
    Include('/tokens', name='tokens', routes=tokens_routes),
    Include('/users', name='users', routes=users_routes),
]



from http.cookies import SimpleCookie
from datetime import datetime, timedelta
from urllib.parse import urlparse


SESSION_COOKIE_NAME = 'session_id'


def get_session_cookie(url: str, session_id: str) -> SimpleCookie:
    parsed = urlparse(url)

    cookie = SimpleCookie()
    cookie[SESSION_COOKIE_NAME] = session_id
    cookie[SESSION_COOKIE_NAME]['path'] = '/'
    cookie[SESSION_COOKIE_NAME]['httponly'] = True

    if parsed.scheme == 'https':
        cookie[SESSION_COOKIE_NAME]['secure'] = True

    cookie[SESSION_COOKIE_NAME]['domain'] = '.' + parsed.netloc

    expires = datetime.utcnow() + timedelta(days=3 * 30)
    date_format = '%a, %d %b %Y %H:%M:%S GMT'
    cookie[SESSION_COOKIE_NAME]['expires'] = expires.strftime(date_format)

    return cookie



from apistar import Route
from apistar.exceptions import HTTPException

from .models import User


def _mark_authorization(f, role=None):
    f.needs_authorization = True
    if role is not None:
        f.authorized_role = role


def authorized(role=None):
    if hasattr(role, '__call__'):
        _mark_authorization(role)
        return role

    def outer(f):
        _mark_authorization(f, role)
        return f
    return outer


class Unauthorized(HTTPException):
    default_status_code = 401
    default_detail = 'Unauthorized'


class AuthorizationHook:
    def on_request(self, route: Route, user: User):
        handler = route.handler
        if not hasattr(handler, 'needs_authorization'):
            return

        if not user:
            raise Unauthorized(dict(error='no authenticated user found'))

        # If there is no authorized role specified, we're done.
        if not hasattr(handler, 'authorized_role'):
            return

        required_role = handler.authorized_role

        if user.role != required_role:
            msg = 'invalid user role "{}" (expected: "{}")'
            error = msg.format(user.role.name, required_role.name)
            raise Unauthorized(dict(error=error))

        # Finally, the request is authorized!



from apistar import Route, validators, types, http
from apistar.exceptions import BadRequest
from sqlalchemy.orm import Session

from .users import User, UserType, UserSession, prune_expired_sessions
from .cookies import get_session_cookie


class LoginType(types.Type):
    username = validators.String(min_length=1)
    password = validators.String(min_length=1)


def login(session: Session, request: http.Request,
          data: LoginType) -> UserType:
    user = session.query(User).filter(User.username == data.username).first()
    if not user:
        raise BadRequest(dict(error='Invalid username/password'))

    verified = user.verify_password(data.password)
    if not verified:
        raise BadRequest(dict(error='Invalid username/password'))

    user_session = UserSession(user=user)
    session.add(user_session)

    cookie = get_session_cookie(request.url, str(user_session.id))
    headers = {'Set-Cookie': cookie.output(header='')}

    prune_expired_sessions(session)

    return http.JSONResponse(UserType(user), headers=headers)


routes = [
    Route('/', 'POST', login),
]



from passlib.hash import bcrypt_sha256, ldap_sha1


# Number of bcrypt rounds (default=12). During testing, bcrypt is not used
# (see also `setup_method()` in ./testutil.py) to speed up the unit tests.
bcrypt_rounds = 12


def enable_bcrypt_hasher(rounds=12):
    global bcrypt_rounds
    bcrypt_rounds = rounds


def disable_bcrypt_hasher():
    global bcrypt_rounds
    bcrypt_rounds = 0


def hasher():
    if bcrypt_rounds:
        return bcrypt_sha256.using(rounds=bcrypt_rounds)
    # During testing, a weaker password hasher is used.
    return ldap_sha1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re

from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with open(os.path.join(package, '__init__.py')) as f:
        init_py = f.read()

    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under root package, that are not in a package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


def get_long_description(long_description_file):
    """
    Read long description from file.
    """
    with open(long_description_file) as f:
        long_description = f.read()

    return long_description


name = 'apistar-auth'
package_name = 'apistar_auth'
version = get_version(package_name)

setup(
    name=name,
    version=version,
    description='Authentication integration based on SQLAlchemy for API Star.',
    long_description=get_long_description('README.rst'),
    author='Sander Mathijs van Veen',
    author_email='sandervv+pypi@gmail.com',
    maintainer='Sander Mathijs van Veen',
    maintainer_email='sandervv+pypi@gmail.com',
    url='https://github.com/smvv/apistar-auth',
    download_url='https://github.com/smvv/apistar-auth',
    packages=get_packages(package_name),
    package_data=get_package_data(package_name),
    install_requires=[
        'apistar',
        'apistar-sqlalchemy',
        'SQLAlchemy',
        'SQLAlchemy-Utils',
        'passlib',
        'bcrypt',
    ],
    tests_require=[
        'apistar',
        'clinner',
        'coverage',
        'prospector',
        'pytest',
        'pytest-xdist',
        'pytest-cov',
        'tox',
    ],
    license='MIT',
    keywords=' '.join([
        'python',
        'apistar',
        'api',
        'sqlalchemy',
        'auth',
        'authentication',
        'login',
        'session',
        'cookie',
        'user',
        'role',
    ]),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from apistar import validators


class UUID(validators.String):
    def validate(self, value, definitions=None, allow_coerce=False):
        return super().validate(str(value), definitions, allow_coerce)



from http.cookies import SimpleCookie
from typing import List
from datetime import datetime, timedelta, timezone
import uuid

from apistar import Route, validators, types, http, Component
from apistar.exceptions import BadRequest
from sqlalchemy.orm import Session
from sqlalchemy.sql.functions import now
from sqlalchemy.exc import IntegrityError

from .auth import authorized
from .cookies import SESSION_COOKIE_NAME
from .models import Token, User, UserRole, UserSession, can_user_create_user
from .validators import UUID


class UserBaseType(types.Type):
    id = validators.Integer(allow_null=True)
    username = validators.String(min_length=1)
    role = validators.String(
        enum=list(UserRole.__members__.keys()),
        allow_null=True,
        default='user'
    )
    fullname = validators.String(min_length=1)

    def __init__(self, *args, **kwargs):
        patched = []
        for arg in args:
            if isinstance(arg, User):
                # Convert enum name to string.
                role = arg.role.name
                arg = arg.__dict__
                arg['role'] = role
            patched.append(arg)
        super().__init__(*patched, **kwargs)


class UserType(UserBaseType):
    created = validators.DateTime()
    updated = validators.DateTime()


class UserInputType(UserBaseType):
    password = validators.String(min_length=1)


class UserSessionType(types.Type):
    id = UUID()
    user_id = validators.Integer()

    created = validators.DateTime()
    updated = validators.DateTime()


session_expires_after = timedelta(days=3 * 30)
session_update_delay = timedelta(days=1)


class UserComponent(Component):
    def __init__(self) -> None:
        pass

    def get_session_id(self, headers):
        cookie_header = headers.get('cookie')
        if not cookie_header:
            return None

        cookie = SimpleCookie()
        cookie.load(cookie_header)

        session_id_cookie = cookie.get(SESSION_COOKIE_NAME)
        if not session_id_cookie:
            return None

        return session_id_cookie.value

    def resolve(self, request: http.Request, session: Session,
                token: http.QueryParam
                # pylint: disable=arguments-differ
                ) -> User:
        if token:
            try:
                token = uuid.UUID(token)
            except ValueError:
                return None

            user = session.query(User) \
                .join(Token) \
                .filter(Token.id == token).first()
            return user

        session_id = self.get_session_id(request.headers)
        if session_id:
            return self.resolve_with_session(session, session_id)

        return None

    def resolve_with_session(self, session, session_id):
        user, session_updated = session.query(User, UserSession.updated) \
            .join(UserSession) \
            .filter(UserSession.id == session_id).first()

        # SQLite does not support datetime timezones. Therefore, it will drop
        # the timezone part. When that's the case, we assume that the timezone
        # was utc since we're comparing with the current utc time.
        if session_updated.tzinfo is None:
            session_updated = session_updated.replace(tzinfo=timezone.utc)

        now_utc = datetime.now(timezone.utc)

        # Reject expired sessions.
        if session_updated <= now_utc - session_expires_after:
            prune_expired_sessions(session)
            session.commit()
            return None

        # Update session field 'updated' when the difference between now and
        # the last update is more than 'session_update_delay'. This avoids
        # updating the field too often.
        if now_utc - session_updated >= session_update_delay:
            session.query(UserSession) \
                .filter(UserSession.id == session_id) \
                .update({'updated': now()}, synchronize_session=False)
            session.commit()

        return user


@authorized(UserRole.admin)
def list_users(session: Session) -> List[UserType]:
    return list(map(UserType, session.query(User).all()))


def create_user(session: Session, user_data: UserInputType,
                user: User) -> http.JSONResponse:
    new_user = User(**dict(user_data))

    if new_user.id is not None:
        raise BadRequest({'error': 'user ID cannot be set'})

    if not can_user_create_user(user, new_user):
        msg = 'user cannot create user with role "{}"'
        raise BadRequest({'error': msg.format(new_user.role.name)})

    session.add(new_user)

    try:
        session.commit()
    except IntegrityError:
        raise BadRequest({'error': 'username already exists'})

    return http.JSONResponse(UserType(new_user), status_code=201)


@authorized
def list_user_session(session: Session, user: User) -> List[UserSessionType]:
    sessions = session.query(UserSession) \
        .filter(UserSession.user_id == user.id) \
        .all()
    return list(map(UserSessionType, sessions))


@authorized
def prune_expired_sessions(session: Session):
    expiration_date = datetime.now(timezone.utc) - session_expires_after
    session.query(UserSession) \
        .filter(UserSession.updated <= expiration_date) \
        .delete(synchronize_session=False)


routes = [
    Route('/', 'GET', list_users),
    Route('/', 'POST', create_user),
    Route('/sessions/', 'GET', list_user_session),
    Route('/sessions/expired/', 'DELETE', prune_expired_sessions),
]
